from typing import Union

from tablate.api.Tablate import Tablate
from tablate.api.modules.Text import Text
from tablate.api.modules.Grid import Grid
from tablate.api.modules.Table import Table
from tablate.classes.bases.TablateApiSet import TablateSet

TablateUnion = Union[Tablate, Text, Grid, Table, TablateSet]


from copy import copy
from typing import List, Union

from tablate.classes.bases.TablateApiSet import TablateApiSet
from tablate.library.calcs.gen_frame_name import gen_frame_name
from tablate.library.calcs.random_string import random_string
from tablate.type.primitives import FrameDivider, Multiline, TextColor, \
    TextAlign, TextStyle, MaxLines, HideHeader, Background, ColumnDivider, ColumnPadding, HtmlPxMultiplier, \
    BackgroundPadding, FrameName
from tablate.type.type_input import GridColumnInput, TableHeaderFrameStylesInput, \
    TableBodyFrameStylesInput, HtmlTableBodyFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    HtmlTableFrameStylesInput, TableColumnInput, \
    TableRowsDataInputDict, \
    HtmlTextFrameStylesInput, \
    HtmlGridFrameStylesInput


class Tablate(TablateApiSet):

    def add_text_frame(self,
                       text: Union[str, int, float],

                       name: FrameName = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       frame_divider: FrameDivider = None,
                       frame_padding: ColumnPadding = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlTextFrameStylesInput = None) -> None:

        # self._frame_list.append({
        #     "name": name,
        #     "type": "text",
        #     "options": locals()
        # })

        # text_frame_store = text_init(text=text,
        #                              name=name,
        #                              frame_divider=frame_divider,
        #                              frame_padding=frame_padding,
        #                              background=background,
        #                              background_padding=background_padding,
        #                              multiline=multiline,
        #                              max_lines=max_lines,
        #                              text_style=text_style,
        #                              text_align=text_align,
        #                              text_color=text_color,
        #                              html_px_multiplier=html_px_multiplier,
        #                              html_styles=html_styles,
        #                              globals=self._globals)

        name = gen_frame_name(name=name, type="text", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]
        args["columns"] = [{"string":text}]

        self._frame_list[name] = {
            "name": name,
            "type": "text",
            "options": args
        }
        # self._frame_list.append(text_frame_store)

    def add_grid_frame(self,
                       columns: List[Union[str, GridColumnInput]],

                       name: FrameName = None,

                       frame_divider: FrameDivider = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       column_divider: ColumnDivider = None,
                       column_padding: ColumnPadding = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlGridFrameStylesInput = None) -> None:

        name = gen_frame_name(name=name, type="grid", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]

        self._frame_list[name] = {
            "name": name,
            "type": "grid",
            "options": args
        }

        # grid_frame_store = grid_init(columns=columns,
        #                              name=name,
        #                              frame_divider=frame_divider,
        #                              background=background,
        #                              background_padding=background_padding,
        #                              multiline=multiline,
        #                              max_lines=max_lines,
        #                              column_divider=column_divider,
        #                              column_padding=column_padding,
        #                              text_style=text_style,
        #                              text_align=text_align,
        #                              text_color=text_color,
        #                              html_styles=html_styles,
        #                              html_px_multiplier=html_px_multiplier,
        #                              globals=self._globals)
        #
        # self._frame_list.append(grid_frame_store)

    def add_table_frame(self,
                        columns: List[TableColumnInput],
                        rows: List[TableRowsDataInputDict],

                        name: FrameName = None,

                        frame_divider: FrameDivider = None,
                        multiline: Multiline = None,
                        max_lines: MaxLines = None,
                        background: Background = None,
                        background_padding: BackgroundPadding = None,

                        multiline_header: Multiline = None,
                        max_lines_header: MaxLines = None,
                        hide_header: HideHeader = None,

                        column_divider: ColumnDivider = None,
                        column_padding: ColumnPadding = None,
                        header_base_divider: FrameDivider = None,

                        row_line_divider: FrameDivider = None,
                        odd_row_background: Background = None,
                        even_row_background: Background = None,

                        text_style: TextStyle = None,
                        text_align: TextAlign = None,
                        text_color: TextColor = None,

                        header_styles: TableHeaderFrameStylesInput = None,
                        body_styles: TableBodyFrameStylesInput = None,

                        html_px_multiplier: HtmlPxMultiplier = None,
                        html_styles: HtmlTableFrameStylesInput = None,

                        html_header_styles: HtmlTableHeaderFrameStylesInput = None,
                        html_body_styles: HtmlTableBodyFrameStylesInput = None) -> None:

        name = gen_frame_name(name=name, type="table", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]

        self._frame_list[name] = {
            "name": name,
            "type": "table",
            "options": args
        }

        # table_header_store, table_body_store = table_init(columns=columns,
        #                                                   rows=rows,
        #                                                   name=name,
        #                                                   frame_divider=frame_divider,
        #                                                   multiline=multiline,
        #                                                   max_lines=max_lines,
        #                                                   background=background,
        #                                                   background_padding=background_padding,
        #                                                   multiline_header=multiline_header,
        #                                                   max_lines_header=max_lines_header,
        #                                                   hide_header=hide_header,
        #                                                   column_divider=column_divider,
        #                                                   column_padding=column_padding,
        #                                                   header_base_divider=header_base_divider,
        #                                                   row_line_divider=row_line_divider,
        #                                                   odd_row_background=odd_row_background,
        #                                                   even_row_background=even_row_background,
        #                                                   text_style=text_style,
        #                                                   text_align=text_align,
        #                                                   text_color=text_color,
        #                                                   header_styles=header_styles,
        #                                                   body_styles=body_styles,
        #                                                   html_px_multiplier=html_px_multiplier,
        #                                                   html_styles=html_styles,
        #                                                   html_header_styles=html_header_styles,
        #                                                   html_body_styles=html_body_styles,
        #                                                   globals=self._globals)
        #
        # if table_header_store:
        #     self._frame_list.append(table_header_store)
        #
        # self._frame_list.append(table_body_store)


########################################################################################################################
########################################################################################################################
########################################################################################################################

#
# column_list: List[Union[GridColumnInput, TableColumnInput]] = [
#     {
#         "string": "Column One",
#         "key": "Key One",
#         "width": "50%",
#         "divider": "thick",
#         "background": "blue"
#     },
#     {
#         "string": "Column Two",
#         "key": "Key Two",
#         "text_align": "right",
#         "divider": "double"
#
#     },
#     {
#         "string": "Column Three",
#         "key": "Key Three",
#         "text_align": "right"
#     }
# ]
#
# column_list2: List[Union[GridColumnInput, TableColumnInput]] = [
#     {
#         "string": "Column One",
#         "key": "Key One",
#         "width": "25%",
#         "divider": "thick",
#         "background": "blue"
#     },
#     {
#         "string": "Column Two",
#         "key": "Key Two",
#         "text_align": "right",
#         "divider": "double"
#
#     },
#     {
#         "string": "Column Three",
#         "key": "Key Three",
#         "text_align": "right"
#     }
# ]
#
# row_list = [
#     {
#         "Key One": "Some string value",
#         "Key Two": 4,
#         "Key Three": 9.6,
#         "divider": "thick"
#     },
#     {
#         "Key One": "Some other string value",
#         "Key Two": 5,
#         "Key Three": 4.6
#
#     },
#     {
#         "Key One": "Some final string value",
#         "Key Two": 100,
#         "Key Three": 328.832
#
#     }
# ]
# ########################################################################################################################
# html_global_styles_input = HtmlOuterBase(html_outer_border_style="thick",
#                                          html_outer_padding=16,
#                                          html_outer_width="100%")
# ########################################################################################################################
# frame_styles_input = FrameStylesInput(frame_divider="thick",
#                                       max_lines=None,
#                                       multiline=True,
#                                       background="yellow",
#                                       trunc_value="...")
# column_styles_input = ColumnStylesInput(divider="thick",
#                                         padding=1)
# text_styles_input = TextStylesInput(text_style="bold",
#                                     text_align="left",
#                                     text_color="green")
# ########################################################################################################################
# html_frame_styles_input = HtmlFrameStylesInput(html_frame_divider_style="thick",
#                                                html_max_lines=5,
#                                                html_multiline=True,
#                                                html_background="green")
# html_column_styles_input = HtmlColumnStylesInput(html_divider_style="thin",
#                                                  html_padding=6)
# html_text_styles_input = HtmlTextStylesInput(html_text_style="normal",
#                                              html_text_align="left",
#                                              html_text_color="red",
#                                              html_text_size=16)
# ########################################################################################################################
# table_rows_frame_styles_input = TableBodyFrameStylesInput(frame_styles=frame_styles_input,
#                                                           column_styles=column_styles_input,
#                                                           text_styles=text_styles_input,
#                                                           row_styles=TableRowsBase(row_line_divider="thin",
#                                                                                    odds_background="grey",
#                                                                                    evens_background="green"))
# header_frame_styles_input = TableHeaderFrameStylesInput(frame_styles=frame_styles_input,
#                                                         column_styles=column_styles_input,
#                                                         text_styles=text_styles_input)
# html_header_styles_input = HtmlTableHeaderFrameStylesInput(html_frame_styles=html_frame_styles_input,
#                                                            html_column_styles=html_column_styles_input,
#                                                            html_text_styles=html_text_styles_input)
# html_rows_styles_input = HtmlTableBodyFrameStylesInput(html_frame_styles=html_frame_styles_input,
#                                                        html_column_styles=html_column_styles_input,
#                                                        html_text_styles=html_text_styles_input,
#                                                        html_row_styles=HtmlTableRowsBase(
#                                                            html_row_line_divider_style="none"
#                                                        ))
# ########################################################################################################################
#
# start = time.perf_counter_ns()
# test = Tablate()
# # test = Tablate(outer_border="thin",
# #                      outer_padding=6,
# #                      outer_width=120,
# #                      html_outer_styles=html_global_styles_input,
# #                      frame_divider="double",
# #                      background="blue",
# #                      column_styles=column_styles_input,
# #                      text_styles=text_styles_input,
# #                      html_frame_styles=html_frame_styles_input,
# #                      html_column_styles=html_column_styles_input,
# #                      html_text_styles=html_text_styles_input)
#
# test.add_text_frame(text="Some String...",
#                     text_style="bold",
#                     text_align="left",
#                     text_color="red",
#                     frame_divider="thick",
#                     background="blue",
#                     multiline=True,
#                     max_lines=5)
#
# test.add_grid_frame(columns=column_list,
#                     frame_divider="thick",
#                     text_color="white",
#                     text_style="bold_underlined",
#                     background="dark_red",
#                     multiline=True,
#                     max_lines=5)
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      odd_row_background="red",
#                      text_color="yellow")
#
# test.add_grid_frame(columns=column_list)
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      name="Some Table",
#                      frame_divider="thick",
#                      background="blue",
#                      multiline=True,
#                      max_lines=5,
#                      multiline_header=False,
#                      max_lines_header=None,
#                      hide_header=False,
#
#                      column_divider="thin",
#                      column_padding=1,
#                      header_base_divider="thick",
#                      row_line_divider="thin",
#                      even_row_background="magenta",
#                      odd_row_background="blue",
#
#                      text_style="bold",
#                      text_align="center",
#                      text_color="blue",
#
#                      header_styles=header_frame_styles_input,
#                      body_styles=table_rows_frame_styles_input,
#
#                      html_styles=html_rows_styles_input,
#
#                      html_header_styles=html_header_styles_input,
#                      html_body_styles=html_rows_styles_input)
#
#
# test.add_text_frame(text="Some String...")
#
# test.add_grid_frame(columns=column_list)
#
# test.add_table_frame(columns=column_list2,
#                      rows=row_list,
#                      hide_header=True)
#
# test.add_grid_frame(columns=column_list, html_px_multiplier=3, text_color="blue")
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      text_color="grey")
#
#
# string = test.to_string()
# html = test.to_html()
# print(string)
# print(html)
#
# test.list_frames()
#
# test.get_frame(9).print()

column_list = [
    {
        "string": "Column One",
        "key": "one",
        "width": "50%",
        "divider": "thick",
    },
    {
        "string": "Column Two",
        "key": "two",
        "text_align": "right",

    },
    {
        "string": "Column Three",
        "key": "three",
        "text_align": "right"
    }
]

row_list = [
    {
        "one": "Some string value",
        "two": 4,
        "three": 9.6,
    },
    {
        "one": "Some other string value",
        "two": 5,
        "three": 4.6

    },
    {
        "one": "Some final string value",
        "two": 100,
        "three": 328.832

    }
]

tab = Tablate(outer_border="thin")

tab.add_table_frame(columns=column_list, rows=row_list, background="red")
tab.add_text_frame("one")
tab.add_table_frame(columns=column_list, rows=row_list, background="green")
tab.add_text_frame("two")
tab.add_table_frame(columns=column_list, rows=row_list, background="blue")
tab.add_text_frame("three")
tab.add_table_frame(columns=column_list, rows=row_list, background="cyan")
tab.add_text_frame("four")
tab.add_table_frame(columns=column_list, rows=row_list, background="yellow")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")





tab.print()
print(tab.to_string())
tab.list_frames()
print(tab.to_html())
tab.get_frame(3).print()
tab.get_frame(3, apply_globals=True).print()
got_frame = tab.get_frame(4, apply_globals=True)
tab.replace_frame(3, got_frame)
tab.remove_frame(0)
tab.print()


from typing import List

from tablate.type.primitives import FrameNameList
from tablate.type.type_store import FrameStoreList
from tablate.type.type_global import Globals


class TablateBase:

    _globals: dict

    _frame_list: dict

    _name_list: FrameNameList



from typing import Callable, List, Union, Optional, Tuple

from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.classes.options.html.style.utilities.base_values import tablate_instance_key
from tablate.classes.options.html.style.utilities.style_dict import style_dict_key_builder, \
    style_dict_css_append
from tablate.classes.options.html.style.utilities.style_types import ElementSelectorDictKeys, SelectorDictUnion, \
    TableSelectorDictKeys, StyleItem
from tablate.library.calcs.random_string import random_string
from tablate.classes.options.html.style.utilities.css_factory import css_factory
from tablate.classes.options.html.style.utilities.selectors import frame_selector, instance_selector, \
    base_selector_dict, table_selector, container_selector


class CssStyler:

    def __init__(self) -> None:
        self._uid = random_string(6)
        self._base_selector = instance_selector(uid=self._uid)
        self.__style_dict: dict = {self._base_selector: {}}
        self.__global_styles = []
        self._css_head = ""
        self._css_foot = ""

    def frame(self, frame_index: int):
        selector_key: ElementSelectorDictKeys = "tablate_frame"
        selector_value = frame_selector(frame_index)
        selector_dict: SelectorDictUnion = base_selector_dict(base_selector=self._base_selector,
                                                              key=selector_key,
                                                              value=selector_value,
                                                              element_type=selector_key)
        return ElementStyler(selector=selector_dict, create_style=self.__create_style())

    @property
    def table(self):
        selector_key: TableSelectorDictKeys = "tablate_container"
        selector_value = table_selector()
        selector_dict: SelectorDictUnion = base_selector_dict(base_selector=self._base_selector,
                                                              key=selector_key,
                                                              value=selector_value,
                                                              element_type=selector_key)
        return ElementStyler(selector=selector_dict, create_style=self.__create_style())

    @property
    def wrapper(self):
        selector_key: TableSelectorDictKeys = "tablate_container"
        selector_value = container_selector()
        selector_dict: SelectorDictUnion = base_selector_dict(base_selector=self._base_selector,
                                                              key=selector_key,
                                                              value=selector_value,
                                                              element_type=selector_key)
        return ElementStyler(selector=selector_dict, create_style=self.__create_style())

    def add_global_style_attribute(self, key: str, value: Union[str, int]):
        self.__global_styles.append(f"{key}:{value}")

    def inject_css_block(self, css: str) -> None:
        self._css_foot += css

    def inject_scoped_css(self, selector: str, css: str) -> None:
        scoped_selector = f"{self._base_selector} {selector}"
        css_block = scoped_selector + "{" + css + "}"
        self._css_foot += css_block

    def return_head_styles(self) -> str:
        return_string = ""
        if len(self.__global_styles) > 0:
            return_string += f".{self._base_selector} *" + "{" + ";".join(self.__global_styles) + ";" + "}"
        return_string += css_factory(self.__style_dict)
        return f"<style>{return_string}</style>"

    def return_foot_styles(self) -> str:
        return f"<style>{self._css_foot}</style>"

    def __create_style(self) -> Callable[[SelectorDictUnion, StyleItem], None]:
        def create_style(selector_dict: SelectorDictUnion, css_item: StyleItem) -> None:

            style_dict_current = self.__style_dict[selector_dict[tablate_instance_key]]
            if "tablate_container" in selector_dict:
                style_dict_current = style_dict_key_builder(style_dict_current, selector_dict, "tablate_container")
            if "tablate_frame" in selector_dict:
                style_dict_current = style_dict_key_builder(style_dict_current, selector_dict, "tablate_frame")
                style_dict_current = style_dict_key_builder(style_dict_current, selector_dict, "tablate_row")
                style_dict_current = style_dict_key_builder(style_dict_current, selector_dict, "tablate_column")
                style_dict_current = style_dict_key_builder(style_dict_current, selector_dict, "tablate_text")
                style_dict_current = style_dict_key_builder(style_dict_current, selector_dict, "element_type")

            style_dict_css_append(style_dict=style_dict_current, css_item=css_item)
        return create_style



from tablate.classes.options.html.style.utilities.base_values import base_class
from tablate.classes.options.html.style.utilities.style_types import SelectorDictUnion


class ClassNameMixin:

    _selector_dict: SelectorDictUnion

    def generate_class_names(self) -> str:
        return_string_array = [base_class]
        for key, value in self._selector_dict.items():
            # return_string_array.append(key)
            if type(value) == str:
                return_string_array.append(value)
        return " ".join(return_string_array)




from tablate.library.renderers.html.frames.render_html_columns import render_html_column
from tablate.library.renderers.html.frames.render_html_table_body import render_html_table_body
from tablate.type.type_store import FrameStoreList
from tablate.type.type_global import Globals


def render_html_frames(frame_list: FrameStoreList, globals: Globals) -> str:

    return_html = ''

    for frame_index, frame_item in enumerate(frame_list):
        frame_item.html_frame_styles.html_px_multiplier = frame_item.html_frame_styles.html_px_multiplier * globals.html.html_outer_styles.html_px_multiplier
        frame_styler = globals.html.styler.frame(frame_index)
        if frame_item.type == "grid" or frame_item.type == "text":
            return_html += render_html_column(frame_dict=frame_item,
                                              globals=globals,
                                              frame_styler=frame_styler)
        if frame_item.type == "table_header":
            return_html += render_html_column(frame_dict=frame_item,
                                              globals=globals,
                                              frame_styler=frame_styler,
                                              frame_type="head")
        if frame_item.type == "table_body":
            return_html += render_html_table_body(table_body_frame_store=frame_item,
                                                  globals=globals,
                                                  frame_styler=frame_styler)

    return return_html



from tablate.library.initializers.mappers.element.base.base_column_mapper import base_column_mapper
from tablate.library.initializers.mappers.element.base.base_frame_mapper import base_frame_mapper
from tablate.library.initializers.mappers.element.base.base_text_mapper import base_text_mapper
from tablate.library.initializers.mappers.element.html.html_column_mapper import html_column_mapper
from tablate.library.initializers.mappers.element.html.html_frame_mapper import html_frame_mapper
from tablate.library.initializers.mappers.element.html.html_outer_mapper import html_outer_mapper
from tablate.library.initializers.mappers.element.html.html_text_mapper import html_text_mapper
from tablate.type.defaults import outer_border_default, outer_padding_default, outer_width_default, \
    background_padding_default, html_px_multiplier_default
from tablate.type.primitives import OuterBorder, OuterPadding, OuterWidth, FrameDivider, Background, BackgroundPadding, \
    HtmlPxMultiplier
from tablate.type.type_base import FrameBase
from tablate.type.type_global import ConsoleGlobals, HtmlGlobals, Globals
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, HtmlFrameStylesInput, \
    HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import OuterStore


def globals_init(outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> Globals:

        default_outer_styles = OuterStore(outer_border=outer_border if outer_border else outer_border_default,
                                          outer_padding=outer_padding if outer_padding else outer_padding_default,
                                          outer_width=outer_width if outer_width else outer_width_default,
                                          background_padding=background_padding if background_padding else background_padding_default)

        default_frame_styles = base_frame_mapper(frame_input=FrameBase(frame_divider=frame_divider,
                                                                       background=background))

        default_column_styles = base_column_mapper(columns_input=column_styles)

        default_text_styles = base_text_mapper(text_input=text_styles)

        console_globals = ConsoleGlobals(outer_styles=default_outer_styles,
                                         frame_styles=default_frame_styles,
                                         column_styles=default_column_styles,
                                         text_styles=default_text_styles)

        if html_px_multiplier is None:
            html_px_multiplier = html_px_multiplier_default

        default_html_outer = html_outer_mapper(html_outer_input=html_styles,
                                               base_outer_defaults=default_outer_styles,
                                               html_px_multiplier=html_px_multiplier)
        default_html_frame = html_frame_mapper(html_frame_input=html_frame_styles,
                                               base_frame_defaults=default_frame_styles)
        default_html_column = html_column_mapper(html_columns_input=html_column_styles,
                                                 base_column_defaults=default_column_styles)
        default_html_text = html_text_mapper(html_text_input=html_text_styles,
                                             base_text_defaults=default_text_styles,
                                             html_px_multiplier=html_px_multiplier)

        html_globals = HtmlGlobals(html_outer_styles=default_html_outer,
                                   html_frame_styles=default_html_frame,
                                   html_column_styles=default_html_column,
                                   html_text_styles=default_html_text,
                                   css_injection="",
                                   column_baselines=[],
                                   styler=None)

        return Globals(console=console_globals, html=html_globals)



from typing import TypedDict, NotRequired, Literal, Union, Optional

TablateInstanceKey = Literal["tablate_instance"]
ElementTypeKey = Literal["element_type"]

BaseSelectorDictKeys = Literal[TablateInstanceKey, ElementTypeKey]
TableSelectorDictKeys = Literal[BaseSelectorDictKeys, "tablate_container"]
ElementSelectorDictKeys = Literal[BaseSelectorDictKeys, "tablate_frame", "tablate_column", "tablate_row", "tablate_text"]

SelectorDictKeysUnion = Union[TableSelectorDictKeys, ElementSelectorDictKeys]

# NOTE: ( ▲ | ▼ ) Ensure the values of these two types match!!! (until Python allows a KeysOf[TypedDict] type...)

class StyleItem(TypedDict):
    style: str
    pseudo: Optional[str]


class BaseSelectorDict(TypedDict):
    tablate_instance: str
    element_type: str


class TableSelectorDict(BaseSelectorDict):
    tablate_table: str


class ElementSelectorDict(BaseSelectorDict):
    tablate_frame: NotRequired[str]
    tablate_column: NotRequired[str]
    tablate_row: NotRequired[str]
    tablate_text: NotRequired[str]


SelectorDictUnion = Union[TableSelectorDict, ElementSelectorDict]



from copy import copy
from typing import List, Union

from tablate.classes.bases.TablateApiSet import TablateApiSet
from tablate.library.calcs.gen_frame_name import gen_frame_name
from tablate.library.calcs.random_string import random_string
from tablate.type.primitives import FrameDivider, Multiline, TextColor, \
    TextAlign, TextStyle, MaxLines, HideHeader, Background, ColumnDivider, ColumnPadding, HtmlPxMultiplier, \
    BackgroundPadding, FrameName
from tablate.type.type_input import GridColumnInput, TableHeaderFrameStylesInput, \
    TableBodyFrameStylesInput, HtmlTableBodyFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    HtmlTableFrameStylesInput, TableColumnInput, \
    TableRowsDataInputDict, \
    HtmlTextFrameStylesInput, \
    HtmlGridFrameStylesInput


class Tablate(TablateApiSet):

    def add_text_frame(self,
                       text: Union[str, int, float],

                       name: FrameName = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       frame_divider: FrameDivider = None,
                       frame_padding: ColumnPadding = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlTextFrameStylesInput = None) -> None:

        name = gen_frame_name(name=name, type="text", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]
        args["columns"] = [{"string":text}]

        self._frame_list[name] = {
            "name": name,
            "type": "text",
            "options": args
        }

    def add_grid_frame(self,
                       columns: List[Union[str, GridColumnInput]],

                       name: FrameName = None,

                       frame_divider: FrameDivider = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       column_divider: ColumnDivider = None,
                       column_padding: ColumnPadding = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlGridFrameStylesInput = None) -> None:

        name = gen_frame_name(name=name, type="grid", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]

        self._frame_list[name] = {
            "name": name,
            "type": "grid",
            "options": args
        }

    def add_table_frame(self,
                        columns: List[TableColumnInput],
                        rows: List[TableRowsDataInputDict],

                        name: FrameName = None,

                        frame_divider: FrameDivider = None,
                        multiline: Multiline = None,
                        max_lines: MaxLines = None,
                        background: Background = None,
                        background_padding: BackgroundPadding = None,

                        multiline_header: Multiline = None,
                        max_lines_header: MaxLines = None,
                        hide_header: HideHeader = None,

                        column_divider: ColumnDivider = None,
                        column_padding: ColumnPadding = None,
                        header_base_divider: FrameDivider = None,

                        row_line_divider: FrameDivider = None,
                        odd_row_background: Background = None,
                        even_row_background: Background = None,

                        text_style: TextStyle = None,
                        text_align: TextAlign = None,
                        text_color: TextColor = None,

                        header_styles: TableHeaderFrameStylesInput = None,
                        body_styles: TableBodyFrameStylesInput = None,

                        html_px_multiplier: HtmlPxMultiplier = None,
                        html_styles: HtmlTableFrameStylesInput = None,

                        html_header_styles: HtmlTableHeaderFrameStylesInput = None,
                        html_body_styles: HtmlTableBodyFrameStylesInput = None) -> None:

        name = gen_frame_name(name=name, type="table", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]

        self._frame_list[name] = {
            "name": name,
            "type": "table",
            "options": args
        }


from dataclasses import dataclass
from typing import TypedDict, NamedTuple
from collections import namedtuple
import attr

from tablate.type.type_base import HtmlOuterBase
from tablate.type.type_input import HtmlOuterStylesInput


class SomeDictClass(TypedDict):
    field3: str

class SomeTuple(NamedTuple):
    field1: str = "bob"
    field2: int = 3

@dataclass
class SomeDataClass:
    field1: str = "bob"
    field2: int = 3


@dataclass
class SomeSubDataClass(SomeDataClass):
    field3: str = ""
    field4: int = 3


some_dataclass = SomeSubDataClass()


# SomeOtherTuple = namedtuple("SomeTuple", ("blah": int))


def some_func(arg1: str, arg2: int, arg3: SomeTuple):
    print(arg3)


some_func(arg1="some string",
          arg2=243,
          arg3=SomeTuple(field1="mary",
                         field2=4))


class SomeSubTuple(SomeTuple):
    field3: str = "bob"
    field4: int = 3

name_tup = SomeSubTuple()

HtmlOuterStylesInput()
HtmlOuterBase()



def gen_frame_name(name: str, type: str, frame_dict: dict):
    if name is not None:
        return name
    else:
        untitled_frame_name = f"Untited{type.capitalize()}Frame0"
        for frame_index, (frame_key, _) in enumerate(frame_dict.items()):
            untitled_frame_name = f"Untited{type.capitalize()}Frame{frame_index}"
            if untitled_frame_name in frame_dict:
                continue
            else:
                return untitled_frame_name
        return untitled_frame_name


from copy import copy

from tablate.library.initializers.mappers.element.base.base_column_mapper import base_column_mapper
from tablate.library.initializers.mappers.element.base.base_frame_mapper import base_frame_mapper
from tablate.library.initializers.mappers.element.base.base_text_mapper import base_text_mapper
from tablate.library.initializers.mappers.element.html.html_column_mapper import html_column_mapper
from tablate.library.initializers.mappers.element.html.html_frame_mapper import html_frame_mapper
from tablate.library.initializers.mappers.element.html.html_outer_mapper import html_outer_mapper
from tablate.library.initializers.mappers.element.html.html_text_mapper import html_text_mapper
from tablate.type.type_base import FrameBase
from tablate.type.type_global import HtmlGlobals

from tablate.classes.bases.TablateBase import TablateBase
from tablate.library.renderers.console.render_console import render_console
from tablate.library.renderers.html.render_html import render_html
from tablate.type.defaults import outer_border_default, outer_padding_default, outer_width_default, \
    html_px_multiplier_default, background_padding_default
from tablate.type.primitives import OuterBorder, FrameDivider, OuterWidth, OuterPadding, Background, HtmlPxMultiplier, \
    BackgroundPadding
from tablate.type.type_global import Globals, ConsoleGlobals
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, \
    HtmlFrameStylesInput, HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import OuterStore


class TablateApiBase(TablateBase):

    def __init__(self,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> None:

        args = copy(locals())
        del args["self"]

        self._globals = args

        # self._name_list = []

        # default_outer_styles = OuterStore(outer_border=outer_border if outer_border else outer_border_default,
        #                                   outer_padding=outer_padding if outer_padding else outer_padding_default,
        #                                   outer_width=outer_width if outer_width else outer_width_default,
        #                                   background_padding=background_padding if background_padding else background_padding_default)
        #
        # default_frame_styles = base_frame_mapper(frame_input=FrameBase(frame_divider=frame_divider,
        #                                                                background=background))
        #
        # default_column_styles = base_column_mapper(columns_input=column_styles)
        #
        # default_text_styles = base_text_mapper(text_input=text_styles)
        #
        # console_globals = ConsoleGlobals(outer_styles=default_outer_styles,
        #                                  frame_styles=default_frame_styles,
        #                                  column_styles=default_column_styles,
        #                                  text_styles=default_text_styles)
        #
        # if html_px_multiplier is None:
        #     html_px_multiplier = html_px_multiplier_default
        #
        # default_html_outer = html_outer_mapper(html_outer_input=html_styles,
        #                                        base_outer_defaults=default_outer_styles,
        #                                        html_px_multiplier=html_px_multiplier)
        # default_html_frame = html_frame_mapper(html_frame_input=html_frame_styles,
        #                                        base_frame_defaults=default_frame_styles)
        # default_html_column = html_column_mapper(html_columns_input=html_column_styles,
        #                                          base_column_defaults=default_column_styles)
        # default_html_text = html_text_mapper(html_text_input=html_text_styles,
        #                                      base_text_defaults=default_text_styles,
        #                                      html_px_multiplier=html_px_multiplier)
        #
        # html_globals = HtmlGlobals(html_outer_styles=default_html_outer,
        #                            html_frame_styles=default_html_frame,
        #                            html_column_styles=default_html_column,
        #                            html_text_styles=default_html_text,
        #                            css_injection="",
        #                            column_baselines=[],
        #                            styler=None)
        #
        # self._globals: Globals = Globals(console=console_globals, html=html_globals)

        self._frame_list = {}

    # def __is_ipython(self):
    #     return hasattr(__builtins__, "__IPYTHON__")

    # def list_frames(self):
    #     print_frame_list(self._frame_list)

    def to_string(self) -> str:
        return render_console(frame_list=self._frame_list, globals=self._globals)

    def print(self) -> None:
        print(self.to_string())

    def __repr__(self) -> str:
        return self.to_string()

    def to_html(self) -> str:
        return render_html(frame_list=self._frame_list, globals=self._globals)

    def _repr_html_(self) -> str:
        return self.to_html()



from typing import overload, TypeVar, Type

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import column_divider_default, column_padding_default, text_style_default, \
    text_align_default, text_color_default, background_default, html_padding_default, html_divider_weight_default, \
    html_px_multiplier_default
from tablate.type.primitives import ColumnDivider, ColumnPadding, TextStyle, TextAlign, TextColor, Background, \
    HtmlPxMultiplier, BackgroundPadding
from tablate.type.type_base import ColumnBase, TextBase, FrameBase
from tablate.type.type_input import BaseColumnInput, GridColumnInput, HtmlColumnInput, \
    TableColumnInput
from tablate.type.type_store import BaseColumnStore


@overload
def column_attr(column_dict: BaseColumnInput,
                frame_styles: FrameBase,
                column_styles: ColumnBase,
                text_styles: TextBase,
                background_padding: BackgroundPadding,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> BaseColumnInput:
    ...


@overload
def column_attr(column_dict: GridColumnInput,
                frame_styles: FrameBase,
                column_styles: ColumnBase,
                text_styles: TextBase,
                background_padding: BackgroundPadding,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> GridColumnInput:
    ...


@overload
def column_attr(column_dict: TableColumnInput,
                frame_styles: FrameBase,
                column_styles: ColumnBase,
                text_styles: TextBase,
                background_padding: BackgroundPadding,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> TableColumnInput:
    ...


T = TypeVar("T", BaseColumnInput, GridColumnInput, TableColumnInput)


def column_attr(column_dict: Type[T],
                frame_styles: FrameBase,
                column_styles: ColumnBase,
                text_styles: TextBase,
                background_padding: BackgroundPadding,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> T:
    divider = set_attr_resolver(instance=column_styles, attr="divider", default=column_divider_default)
    padding = set_attr_resolver(instance=column_styles, attr="padding", default=column_padding_default)
    text_style = set_attr_resolver(instance=text_styles, attr="text_style", default=text_style_default)
    text_align = set_attr_resolver(instance=text_styles, attr="text_align", default=text_align_default)
    text_color = set_attr_resolver(instance=text_styles, attr="text_color", default=text_color_default)
    background = set_attr_resolver(instance=frame_styles, attr="background", default=background_default)
    html_styles = set_attr_resolver(instance=frame_styles, attr="html_styles", default={})

    if "divider" in column_dict and column_dict["divider"] is not None:
        divider = column_dict["divider"]
        html_styles["divider_style"] = column_dict["divider"]
        html_styles["divider_weight"] = html_divider_weight_default * html_px_multiplier
    if "padding" in column_dict and column_dict["padding"] is not None:
        padding = column_dict["padding"]
        html_styles["padding"] = column_dict["padding"] * html_padding_default
    if "text_style" in column_dict and column_dict["text_style"] is not None:
        text_style = column_dict["text_style"]
        html_styles["text_style"] = column_dict["text_style"]
    if "text_align" in column_dict and column_dict["text_align"] is not None:
        text_align = column_dict["text_align"]
        html_styles["text_align"] = column_dict["text_align"]
    if "text_color" in column_dict and column_dict["text_color"] is not None:
        text_color = column_dict["text_color"]
        html_styles["text_color"] = column_dict["text_color"]
    if "background" in column_dict and column_dict["background"] is not None:
        background = column_dict["background"]
        html_styles["background"] = column_dict["background"]

    if "html_styles" in column_dict or ("html_styles" in column_dict and column_dict["html_styles"] is None):
        html_styles = None

    column_dict: BaseColumnStore = {
        "width": column_dict["width"],
        "string": column_dict["string"],
        "divider": divider,
        "padding": padding,
        "text_style": text_style,
        "text_align": text_align,
        "text_color": text_color,
        "background": background,
        "html_styles": html_styles if html_styles is not None else column_dict["html_styles"],
        "background_padding": background_padding,
        **column_dict
    }
    return_dict = BaseColumnStore(**column_dict)

    return return_dict


# todo: can use TypeVar + Type[T] (hopefully) to do away with the overloads... or not...

# todo: fix this!



from copy import copy
from typing import Union

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.classes.bases.TablateItem import TablateItem
from tablate.library.calcs.gen_frame_name import gen_frame_name
from tablate.library.initializers import text_init
from tablate.type.primitives import TextAlign, FrameDivider, OuterBorder, FrameName, TextStyle, TextColor, Background, \
    BackgroundPadding, Multiline, MaxLines, HtmlPxMultiplier, OuterPadding, OuterWidth, ColumnPadding
from tablate.type.type_input import HtmlTextFrameStylesInput, HtmlOuterStylesInput


class Text(TablateApiItem):

    def __init__(self,
                 # TablateText args
                 text: Union[str, int, float],

                 name: FrameName = None,

                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,
                 frame_padding: ColumnPadding = None,
                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlTextFrameStylesInput = None,
                 # TablateApi arge
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,
                 html_outer_styles: HtmlOuterStylesInput = None) -> None:

        TablateApiBase.__init__(self=self,
                                outer_border=outer_border,
                                outer_padding=outer_padding,
                                frame_divider=frame_divider,
                                outer_width=outer_width,
                                html_styles=html_outer_styles)

        name = gen_frame_name(name=name, type="text", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]
        args["columns"] = [{"string": text}]

        self._frame_list[name] = {
            "name": name,
            "type": "text",
            "options": args
        }


from typing import Union, Callable, Tuple

from tablate.classes.bases.TablateApiBase import TablateApiBase


class TablateApiItem(TablateApiBase):

    def to_dict(self):
        pass

    def apply(self, function: Callable[[dict, dict], Tuple[dict, dict]]):
        frame_dict, globals_dict = function(self._frame_list[0], self._globals)
        if frame_dict is not None:
            self._frame_list[0] = frame_dict
        if globals_dict is not None:
            self._globals = globals_dict


import copy
from typing import List, Union, Any

from tablate.library.calcs.calc_column_widths import calc_column_widths
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.initializers.mappers.element.base.base_frame_mapper import base_frame_mapper
from tablate.library.initializers.mappers.element.base.base_rows_mapper import base_rows_mapper
from tablate.library.initializers.mappers.element.html.html_rows_mapper import html_rows_mapper
from tablate.library.initializers.mappers.element.style_mapper import style_mapper
from tablate.library.initializers.mappers.attribute.column_attr import column_attr
from tablate.type.defaults import table_multiline_default, table_header_text_style_default, header_base_divider_default, \
    background_padding_default, frame_name_default
from tablate.type.type_global import Globals
from tablate.type.type_store import TableBodyFrameStore, TableHeaderFrameStore, FrameStore, ColumnStore, TextStore, \
    TableRowsStore, HtmlTableRowsStore
from tablate.type.type_input import TableRowsDataInputDict, TableHeaderFrameStylesInput, TableBodyFrameStylesInput, \
    HtmlTableFrameStylesInput, HtmlTableHeaderFrameStylesInput, HtmlTableBodyFrameStylesInput, TableColumnInput, \
    BaseStylesInput, HtmlStylesInput
from tablate.type.primitives import FrameDivider, ColumnDivider, Multiline, MaxLines, HideHeader, ColumnPadding, \
    TextStyle, TextAlign, TextColor, Background, HtmlPxMultiplier, BackgroundPadding, FrameName


def table_init(columns: List[TableColumnInput],
               rows: List[TableRowsDataInputDict],

               name: FrameName,

               frame_divider: FrameDivider,
               multiline: Multiline,
               max_lines: MaxLines,
               background: Background,
               background_padding: BackgroundPadding,

               multiline_header: Multiline,
               max_lines_header: MaxLines,
               hide_header: HideHeader,

               column_divider: ColumnDivider,
               column_padding: ColumnPadding,
               header_base_divider: FrameDivider,

               row_line_divider: FrameDivider,
               odd_row_background: Background,
               even_row_background: Background,

               text_style: TextStyle,
               text_align: TextAlign,
               text_color: TextColor,

               header_styles: TableHeaderFrameStylesInput,
               body_styles: TableBodyFrameStylesInput,

               html_px_multiplier: HtmlPxMultiplier,
               html_styles: HtmlTableFrameStylesInput,

               html_header_styles: HtmlTableHeaderFrameStylesInput,
               html_body_styles: HtmlTableBodyFrameStylesInput,

               globals: Globals) -> (Union[TableHeaderFrameStore, None], TableBodyFrameStore):

    ####################################################################################################################

    columns = copy.deepcopy(columns)
    rows = copy.deepcopy(rows)

    ####################################################################################################################

    background_padding = background_padding if background_padding is not None else set_attr_resolver(
        instance=globals.console.outer_styles,
        attr="background_padding",
        default=background_padding_default)

    name = name if name is not None else frame_name_default
    hide_header = hide_header if hide_header is not None else False

    #############################################
    #############################################
    #############################################

    table_all_frame_defaults = base_frame_mapper(frame_input=globals.console.frame_styles)
    table_all_frame_defaults.multiline = multiline if multiline else table_multiline_default

    table_all_base_defaults = BaseStylesInput(frame_styles=base_frame_mapper(frame_input=globals.console.frame_styles),
                                         column_styles=globals.console.column_styles,
                                         text_styles=globals.console.text_styles)

    table_all_html_defaults = HtmlStylesInput(html_frame_styles=globals.html.html_frame_styles,
                                              html_column_styles=globals.html.html_column_styles,
                                              html_text_styles=globals.html.html_text_styles)

    #############################################

    table_styles = style_mapper(base_input=BaseStylesInput(frame_styles=FrameStore(frame_divider=frame_divider,
                                                                                   max_lines=max_lines,
                                                                                   multiline=multiline,
                                                                                   background=background),
                                                           column_styles=ColumnStore(divider=column_divider,
                                                                                     padding=column_padding,
                                                                                     background_padding=background_padding),
                                                           text_styles=TextStore(text_style=text_style,
                                                                                 text_align=text_align,
                                                                                 text_color=text_color)),
                                base_defaults=table_all_base_defaults,
                                html_input=html_styles,
                                html_defaults=table_all_html_defaults,
                                html_px_multiplier=html_px_multiplier,
                                globals=globals)

    ####################################################################################################################

    header_base_frame_defaults = copy.deepcopy(table_styles.frame_styles)
    header_base_frame_defaults.frame_divider = header_base_divider if header_base_divider else header_base_divider_default
    header_base_frame_defaults.multiline = multiline_header
    header_base_frame_defaults.max_lines = max_lines_header

    header_base_text_defaults = copy.deepcopy(table_styles.text_styles)
    header_base_text_defaults.text_style = table_header_text_style_default

    compiled_header_styles = style_mapper(base_input=header_styles,
                                          html_input=html_header_styles,
                                          base_defaults=BaseStylesInput(frame_styles=header_base_frame_defaults,
                                                                       column_styles=table_styles.column_styles,
                                                                       text_styles=header_base_text_defaults),
                                          html_defaults=table_all_html_defaults,
                                          html_px_multiplier=html_px_multiplier,
                                          globals=globals)

    ####################################################################################################################

    compiled_body_styles = style_mapper(base_input=body_styles,
                                        html_input=html_body_styles,
                                        base_defaults=BaseStylesInput(frame_styles=table_styles.frame_styles,
                                                                      column_styles=table_styles.column_styles,
                                                                      text_styles=table_styles.text_styles),
                                        html_defaults=table_all_html_defaults,
                                        html_px_multiplier=html_px_multiplier,
                                        globals=globals)

    html_body_rows_styles = set_attr_resolver(html_body_styles, "html_row_styles", HtmlTableRowsStore())

    body_base_row_styles = base_rows_mapper(rows_input=TableRowsStore(row_line_divider=row_line_divider,
                                                                      evens_background=even_row_background,
                                                                      odds_background=odd_row_background))
    if hasattr(body_styles, "row_styles"):
        body_base_row_styles = base_rows_mapper(
            rows_input=TableRowsStore(row_line_divider=set_attr_resolver(instance=body_styles.row_styles,
                                                                         attr="row_line_divider",
                                                                         default=body_base_row_styles.row_line_divider),
                                      evens_background=set_attr_resolver(instance=body_styles.row_styles,
                                                                              attr="evens_background",
                                                                              default=body_base_row_styles.evens_background),
                                      odds_background=set_attr_resolver(instance=body_styles.row_styles,
                                                                             attr="odds_background",
                                                                             default=body_base_row_styles.odds_background)))
    body_html_row_styles = html_rows_mapper(html_rows_input=html_body_rows_styles,
                                            base_rows_defaults=body_base_row_styles)

    #############################################
    #############################################
    #############################################

    columns = calc_column_widths(columns=columns, globals=globals)

    header_column_list = []
    rows_column_list = []

    for column_item in columns:
        column_item["string"] = column_item["string"] if "string" in column_item else column_item["key"]

        head_column_dict = column_attr(column_dict=column_item,
                                       frame_styles=compiled_header_styles.frame_styles,
                                       column_styles=compiled_header_styles.column_styles,
                                       text_styles=compiled_header_styles.text_styles,
                                       background_padding=background_padding)

        body_column_dict = column_attr(column_dict=column_item,
                                       frame_styles=compiled_body_styles.frame_styles,
                                       column_styles=compiled_body_styles.column_styles,
                                       text_styles=compiled_body_styles.text_styles,
                                       background_padding=background_padding)

        header_column_list.append(head_column_dict)

        rows_column_list.append(body_column_dict)

    header_frame_store: Union[TableHeaderFrameStore, None]

    if not hide_header:
        header_frame_store = TableHeaderFrameStore(type="table_header",
                                                   name=name,
                                                   column_list=header_column_list,
                                                   frame_styles=compiled_header_styles.frame_styles,
                                                   column_styles=compiled_header_styles.column_styles,
                                                   text_styles=compiled_header_styles.text_styles,
                                                   html_frame_styles=compiled_header_styles.html_frame_styles,
                                                   html_column_styles=compiled_header_styles.html_column_styles,
                                                   html_text_styles=compiled_header_styles.html_text_styles)
    else:
        header_frame_store = None

    body_frame_store = TableBodyFrameStore(type="table_body",
                                           name=name,
                                           hide_header=hide_header,
                                           column_list=rows_column_list,
                                           row_list=rows,
                                           frame_styles=compiled_body_styles.frame_styles,
                                           row_styles=body_base_row_styles,
                                           column_styles=compiled_body_styles.column_styles,
                                           text_styles=compiled_body_styles.text_styles,
                                           html_frame_styles=compiled_body_styles.html_frame_styles,
                                           html_row_styles=body_html_row_styles,
                                           html_column_styles=compiled_body_styles.html_column_styles,
                                           html_text_styles=compiled_body_styles.html_text_styles)

    return header_frame_store, body_frame_store



from tablate.classes.options.html.style.utilities.base_values import base_class, tablate_instance_key, \
    element_type_key
from tablate.classes.options.html.style.utilities.style_types import ElementSelectorDictKeys


def element_append(string: str) -> str:
    return f"{string}_element"


def base_selector_dict(base_selector: str, key: str, value: str, element_type: ElementSelectorDictKeys):
    # element_type = element_append(element_type)
    return {tablate_instance_key: base_selector, key: value, element_type_key: element_type}


def instance_selector(uid: str):
    return f"{base_class}_{uid}"


def container_selector() -> str:
    return f"{base_class}_wrapper"


def table_selector() -> str:
    return f"{base_class}_table"


def frame_selector(index: int) -> str:
    return f"{base_class}_frame_{index}"


def column_selector(index: int) -> str:
    specifier = ""
    if index is not None:
        specifier = f"_{index}"
    return f"{base_class}_column{specifier}"


def row_selector(index: int) -> str:
    specifier = ""
    if index is not None:
        specifier = f"_{index}"
    return f"{base_class}_row{specifier}"


def text_selector() -> str:
    return f"{base_class}_text"



from typing import Union, Callable, Tuple

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.helpers.get_frame import get_frame
from tablate.classes.helpers.list_frame import list_frames


class TablateApiSet(TablateApiBase):

    def apply(self, function: Callable[[dict, dict],Tuple[dict, dict]]):
        frame_list, globals_dict = function(self._frame_list, self._globals)
        if frame_list is not None:
            self._frame_list = frame_list
        if globals_dict is not None:
            self._globals = globals_dict

    def list_frames(self):
        list_frames(self._frame_list)

    def get_frame(self, selector: Union[int, str], apply_globals: bool = False):
        if apply_globals:
            return get_frame(frame_list=self._frame_list, selector=selector, globals=self._globals)
        else:
            return get_frame(frame_list=self._frame_list, selector=selector)

    def from_dict(self):
        pass

    def remove_frame(self, selector: Union[int, str]):
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if (type(selector) == int and selector == frame_index) or (type(selector) == str and selector == frame_key):
                del self._frame_list[frame_key]
                break

    def replace_frame(self, selector: Union[int, str], new_frame):
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if (type(selector) == int and selector == frame_index) or (type(selector) == str and selector == frame_key):
                for new_frame_key, new_frame_item in new_frame._frame_list.items():
                    self._frame_list[frame_key] = new_frame_item
                    break


from tablate.classes.options.html.style.subclasses.TextStyler import TextStyler
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.html.style.attributes.color import text_color_attr
from tablate.type.defaults import html_text_size_default
from tablate.type.primitives import HtmlPxMultiplier
from tablate.type.type_base import HtmlTextBase


def style_text(text_store: HtmlTextBase, text_styler: TextStyler, html_px_multiplier: HtmlPxMultiplier) -> None:
    text_align = text_store.html_text_align
    text_size = set_attr_resolver(instance=text_store, attr="html_text_size", default=html_text_size_default)
    text_style = text_store.html_text_style
    text_color = text_store.html_text_color

    text_styler.add_style_attribute("font-size", f"{text_size * html_px_multiplier}px")

    if text_align is not None:
        text_styler.add_style_attribute("text-align", text_align)
    if text_style is not None:
        if text_style == "bold" or text_style == "bold_underlined":
            text_styler.add_style_attribute("font-weight", "bold")
        if text_style == "underlined" or text_style == "bold_underlined":
            text_styler.add_style_attribute("text-decoration", "underline")
    if text_color is not None:
        text_styler.add_style_attribute("color", text_color_attr(text_color))





from tablate.library.ascii.chars.line_v import v_line
from tablate.library.ascii.chars.matrix_side import left_side_matrix, right_side_matrix
from tablate.type.primitives import FrameDivider
from tablate.type.type_global import Globals


def row_outer_border(row_string: str, globals: Globals, row_divider: FrameDivider = None) -> str:

    outer_padding = globals.console.outer_styles.outer_padding
    outer_border = globals.console.outer_styles.outer_border

    left_border = left_side_matrix[outer_border][row_divider] if row_divider else v_line[outer_border]
    right_border = right_side_matrix[outer_border][row_divider] if row_divider else v_line[outer_border]

    grid_line_inner = f"{left_border}{row_string}{right_border}"
    return_string = f"{' ' * outer_padding}{grid_line_inner}\n"

    return return_string



from dataclasses import dataclass, field
from typing import List, Optional

from tablate.classes.options.html.style.CssStyler import CssStyler
from tablate.type.primitives import CssStyleBlock, PxInteger
from tablate.type.type_base import HtmlOuterBase, FrameBase, ColumnBase, TextBase, OuterBase, HtmlFrameBase, \
    HtmlColumnBase, HtmlTextBase
from tablate.type.type_store import HtmlOuterStore, OuterStore


@dataclass
class ConsoleGlobals:
    outer_styles: OuterStore = field(default_factory=OuterStore)
    frame_styles: FrameBase = field(default_factory=FrameBase)
    column_styles: ColumnBase = field(default_factory=ColumnBase)
    text_styles: TextBase = field(default_factory=TextBase)


@dataclass
class HtmlGlobals:
    html_outer_styles: HtmlOuterStore = field(default_factory=HtmlOuterStore)
    html_frame_styles: HtmlFrameBase = field(default_factory=HtmlFrameBase)
    html_column_styles: HtmlColumnBase = field(default_factory=HtmlColumnBase)
    html_text_styles: HtmlTextBase = field(default_factory=HtmlTextBase)
    css_injection: CssStyleBlock = ""
    styler: Optional[CssStyler] = field(default_factory=CssStyler)
    column_baselines: List[PxInteger] = field(default_factory=list[6])


@dataclass
class Globals:
    console: ConsoleGlobals = field(default_factory=ConsoleGlobals)
    html: HtmlGlobals = field(default_factory=HtmlGlobals)



from typing import Optional

from tablate.type.primitives import HtmlSpacer, HtmlPxMultiplier


def space_attr(html_spacer: Optional[HtmlSpacer], html_px_multiplier: HtmlPxMultiplier):
    space_css = "0"
    edge_space = 0
    if html_spacer is not None:
        if type(html_spacer) == int or type(html_spacer) == str:
            space_css = f"{html_spacer * html_px_multiplier}px"
            edge_space = int(html_spacer) * 2
        if type(html_spacer) == list:
            if len(html_spacer) == 1:
                space_css = f"{html_spacer[0] * html_px_multiplier}px"
                edge_space = int(html_spacer[0]) * 2
            if len(html_spacer) == 2:
                space_css = f"{html_spacer[0] * html_px_multiplier}px {html_spacer[1] * html_px_multiplier}px"
                edge_space = int(html_spacer[1]) * 2
            if len(html_spacer) == 4:
                space_css = f"{html_spacer[0] * html_px_multiplier}px {html_spacer[1] * html_px_multiplier}px {html_spacer[2] * html_px_multiplier}px {html_spacer[3] * html_px_multiplier}px"
                edge_space = int(html_spacer[1]) + int(html_spacer[3])
    return space_css, edge_space



import copy
from typing import List, Union, Any

from tablate.library.calcs.calc_column_widths import calc_column_widths
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.initializers.mappers.element.style_mapper import style_mapper
from tablate.library.initializers.mappers.attribute.column_attr import column_attr
from tablate.type.defaults import background_padding_default, frame_name_default
from tablate.type.type_global import Globals
from tablate.type.type_store import GridFrameStore, GridColumnStore, FrameStore, ColumnStore, TextStore
from tablate.type.type_input import GridColumnInput, BaseStylesInput, HtmlGridFrameStylesInput, HtmlStylesInput
from tablate.type.primitives import FrameDivider, ColumnDivider, Background, Multiline, MaxLines, ColumnPadding, \
    TextStyle, TextAlign, TextColor, HtmlPxMultiplier, BackgroundPadding, FrameName


def grid_init(columns: List[Union[str, GridColumnInput]],
              name: FrameName,
              frame_divider: FrameDivider,
              background: Background,
              background_padding: BackgroundPadding,
              multiline: Multiline,
              max_lines: MaxLines,

              column_divider: ColumnDivider,
              column_padding: ColumnPadding,

              text_style: TextStyle,
              text_align: TextAlign,
              text_color: TextColor,

              html_px_multiplier: HtmlPxMultiplier,
              html_styles: HtmlGridFrameStylesInput,

              globals: Globals) -> GridFrameStore:
    columns = copy.deepcopy(columns)

    background_padding = background_padding if background_padding is not None else set_attr_resolver(
        instance=globals.console.outer_styles,
        attr="background_padding",
        default=background_padding_default)

    name = name if name is not None else frame_name_default

    grid_styles = style_mapper(base_input=BaseStylesInput(frame_styles=FrameStore(frame_divider=frame_divider,
                                                                                  max_lines=max_lines,
                                                                                  multiline=multiline,
                                                                                  background=background),
                                                          column_styles=ColumnStore(divider=column_divider,
                                                                                    padding=column_padding,
                                                                                    background_padding=background_padding),
                                                          text_styles=TextStore(text_style=text_style,
                                                                                text_align=text_align,
                                                                                text_color=text_color)),
                               html_input=html_styles,
                               base_defaults=BaseStylesInput(),
                               html_defaults=HtmlStylesInput(),
                               html_px_multiplier=html_px_multiplier,
                               globals=globals)

    columns = calc_column_widths(columns=columns, globals=globals)

    grid_column_list: List[GridColumnStore] = []

    for column_item in columns:

        if type(column_item) == str:
            column_item = {"string": column_item}

        grid_column_dict = column_attr(column_dict=column_item,
                                       frame_styles=grid_styles.frame_styles,
                                       column_styles=grid_styles.column_styles,
                                       text_styles=grid_styles.text_styles,
                                       background_padding=background_padding)
        grid_column_list.append(grid_column_dict)

    grid_frame_store = GridFrameStore(type="grid",
                                      name=name,
                                      column_list=grid_column_list,
                                      frame_styles=grid_styles.frame_styles,
                                      column_styles=grid_styles.column_styles,
                                      text_styles=grid_styles.text_styles,
                                      html_frame_styles=grid_styles.html_frame_styles,
                                      html_column_styles=grid_styles.html_column_styles,
                                      html_text_styles=grid_styles.html_text_styles)

    return grid_frame_store



from typing import List

from tablate.library.ascii.chars.line_h import h_line
from tablate.library.ascii.chars.matrix_cross import cross_matrix
from tablate.type.primitives import FrameDivider
from tablate.type.type_store import BaseColumnStore


def row_line_divider(column_list_top: List[BaseColumnStore], column_list_bottom: List[BaseColumnStore], divider: FrameDivider) -> str:

    # todo: fix this!
    #  are both the 'col_width' and the 'sum_width' needed ??
    #  (sum_width is used for if checks... col_width are used for iterative calculations..)

    return_string = ""

    upper_frame_sum_width = 0
    upper_frame = []
    for v_line_index in range(0, len(column_list_top)):
        upper_frame_sum_width += column_list_top[v_line_index]["width"]
        upper_frame.append({
            "col_width": column_list_top[v_line_index]["width"],
            "sum_width": upper_frame_sum_width,
            "divider": column_list_top[v_line_index]["divider"]
        })
    upper_frame.reverse()
    
    lower_frame_sum_width = 0
    lower_frame = []
    for v_line_index in range(0, len(column_list_bottom)):
        lower_frame_sum_width += column_list_bottom[v_line_index]["width"]
        lower_frame.append({
            "col_width": column_list_bottom[v_line_index]["width"], 
            "sum_width": lower_frame_sum_width,              
            "divider": column_list_bottom[v_line_index]["divider"]
        })
    lower_frame.reverse()

    current_v_line_index = len(column_list_top) + len(column_list_bottom)

    while current_v_line_index > 0:
        if len(upper_frame) > 0 and len(lower_frame) > 0:
            if upper_frame[-1]["sum_width"] > lower_frame[-1]["sum_width"]:
                return_string += f"{h_line[divider] * lower_frame[-1]['col_width']}{cross_matrix['blank'][divider][lower_frame[-1]['divider']]}"
                upper_frame[-1]["col_width"] = upper_frame[-1]["col_width"] - lower_frame[-1]["col_width"] - 1
                lower_frame.pop()
            elif upper_frame[-1]["sum_width"] == lower_frame[-1]["sum_width"]:
                return_string += f"{h_line[divider] * lower_frame[-1]['col_width']}{cross_matrix[upper_frame[-1]['divider']][divider][lower_frame[-1]['divider']]}"
                upper_frame.pop()
                lower_frame.pop()
            elif upper_frame[-1]["sum_width"] < lower_frame[-1]["sum_width"]:
                return_string += f"{h_line[divider] * upper_frame[-1]['col_width']}{cross_matrix[upper_frame[-1]['divider']][divider]['blank']}"
                lower_frame[-1]["col_width"] = lower_frame[-1]["col_width"] - upper_frame[-1]["col_width"] - 1
                upper_frame.pop()
        else:
            if len(upper_frame) > 0:
                return_string += f"{h_line[divider] * upper_frame[-1]['col_width']}{cross_matrix[upper_frame[-1]['divider']][divider]['blank']}"
                if upper_frame[-1]["col_width"] < 0:
                    return_string = return_string[0:upper_frame[-1]["col_width"]]
                upper_frame.pop()
            if len(lower_frame) > 0:
                return_string += f"{h_line[divider] * lower_frame[-1]['col_width']}{cross_matrix['blank'][divider][lower_frame[-1]['divider']]}"
                if lower_frame[-1]["col_width"] < 0:
                    return_string = return_string[0:lower_frame[-1]["col_width"]]
                lower_frame.pop()
        current_v_line_index -= 1

    return return_string[0:-1]



from tablate.library.ascii.chars.corners import bottom_left, bottom_right
from tablate.library.formatters.console.row_line_divider import row_line_divider
from tablate.type.type_store import FrameStoreList
from tablate.type.type_global import Globals


def render_console_foot(frame_list: FrameStoreList, globals: Globals) -> str:

    return_string = ""

    outer_padding = globals.console.outer_styles.outer_padding
    frame_border = globals.console.outer_styles.outer_border

    bottom_border_inner = row_line_divider(column_list_top=frame_list[-1].column_list,
                                           column_list_bottom=[],
                                           divider=frame_border)
    return_string += f"{' ' * outer_padding}{bottom_left[frame_border]}{bottom_border_inner}{bottom_right[frame_border]}\n"

    return return_string


from typing import Optional

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import column_divider_default, background_default, column_padding_default
from tablate.type.type_base import ColumnBase
from tablate.type.type_store import ColumnStore


def base_column_mapper(columns_input: Optional[ColumnBase] = None,
                       column_defaults: Optional[ColumnBase] = None) -> ColumnStore:

    ####################################################################################################################

    columns_input = columns_input if columns_input is not None else ColumnBase()
    column_defaults = column_defaults if column_defaults is not None else ColumnBase()

    ####################################################################################################################

    columns_return = ColumnStore(divider=set_attr_resolver(instance=columns_input,
                                                           attr="divider",
                                                           default=set_attr_resolver(instance=column_defaults,
                                                                                     attr="divider",
                                                                                     default=column_divider_default)),
                                 padding=set_attr_resolver(instance=columns_input,
                                                           attr="padding",
                                                           default=set_attr_resolver(instance=column_defaults,
                                                                                     attr="padding",
                                                                                     default=column_padding_default)))

    return columns_return



from copy import copy
from typing import List

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.library.calcs.gen_frame_name import gen_frame_name
from tablate.type.primitives import FrameDivider, ColumnDivider, OuterBorder, \
    FrameName, Multiline, MaxLines, Background, BackgroundPadding, HideHeader, ColumnPadding, TextAlign, TextStyle, \
    TextColor, HtmlPxMultiplier, OuterPadding, OuterWidth
from tablate.type.type_input import TableColumnInput, TableRowsDataInputDict, TableBodyFrameStylesInput, \
    TableHeaderFrameStylesInput, HtmlTableFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    HtmlTableBodyFrameStylesInput, HtmlOuterStylesInput


class Table(TablateApiItem):

    def __init__(self,
                 # TablateTable args
                 columns: List[TableColumnInput],
                 rows: List[TableRowsDataInputDict],

                 name: FrameName = None,

                 frame_divider: FrameDivider = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 multiline_header: Multiline = None,
                 max_lines_header: MaxLines = None,
                 hide_header: HideHeader = None,

                 column_divider: ColumnDivider = None,
                 column_padding: ColumnPadding = None,
                 header_base_divider: FrameDivider = None,

                 row_line_divider: FrameDivider = None,
                 odd_row_background: Background = None,
                 even_row_background: Background = None,

                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,

                 header_styles: TableHeaderFrameStylesInput = None,
                 body_styles: TableBodyFrameStylesInput = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlTableFrameStylesInput = None,

                 html_header_styles: HtmlTableHeaderFrameStylesInput = None,
                 html_body_styles: HtmlTableBodyFrameStylesInput = None,

                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,
                 html_outer_styles: HtmlOuterStylesInput = None) -> None:

        TablateApiBase.__init__(self=self,
                                outer_border=outer_border,
                                outer_padding=outer_padding,
                                frame_divider=frame_divider,
                                outer_width=outer_width,
                                html_styles=html_outer_styles)

        name = gen_frame_name(name=name, type="table", frame_dict=self._frame_list)

        args = copy(locals())
        del args["self"]

        self._frame_list[name] = {
            "name": name,
            "type": "table",
            "options": args
        }